import 'dart:async';
import 'dart:convert' as convert;
import 'dart:convert';

import 'package:apex_flutter_plugin/api/config.dart';
import 'package:connectivity/connectivity.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:http/http.dart' as http;
import 'package:http_parser/http_parser.dart';

import 'base_request.dart';
import 'cipher/crypto.dart';
import 'multipart_request.dart';
import 'request.dart';
import 'response.dart';

class Api {
  ApiConfig config = ApiConfig('https://google.com', '0', '0');
  http.Client? client = http.Client();

  static final Api _singleton = Api._();

  Api();

  Api._();

  static Api get shared => _singleton;

//  isolateRequest(IsolateModel isolateModel) {
//    compute(makeRequest, isolateModel);
//  }
//
//  Future loadIsolate(
//      IsolateModel isolateModel, Function(Response) printFunc) async {
//    ReceivePort receivePort = ReceivePort();
//
//    await Isolate.spawn(isolateEntry, receivePort.sendPort);
//
//    SendPort sendPort = await receivePort.first;
//
//    Response response = await sendReceive(sendPort,
//        isolateModel); //todo send message to get Response object in return
//
////    isolate.kill(); //todo test
//
//    printFunc(response);
//
//    //todo remove return
//    return response;
//  }
//
//  static isolateEntry(SendPort sendPort) async {
//    ReceivePort port = ReceivePort();
//
//    sendPort.send(port.sendPort);
//
//    await for (var msg in port) {
//      SendPort replyPort = msg[0];
//      Request request = msg[1];
//      String url = msg[2];
//      convert.Encoding encoding = msg[3];
//      Map<String, String> headers = msg[4];
//      Duration timeLimit = msg[5];
//      OnTimeout onTimeout = msg[6];
//
//      Response response = await Api.shared.makeRequest(IsolateModel(request,
//          url: url,
//          encoding: encoding,
//          headers: headers,
//          timeLimit: timeLimit,
//          onTimeout: onTimeout));
//
//      replyPort.send(response); //send the reply back to sendReceive
//    }
//  }
//
//  Future sendReceive(SendPort send, IsolateModel isolateModel) {
//    ReceivePort responsePort = ReceivePort();
//
//    send.send([responsePort.sendPort, isolateModel]); //sending to isolateEntry
//
//    return responsePort.first;
//  }

  String _prepareRequestMessage(Crypto crypto, IsolateModel isolateModel) {
    _print('default parameters : ' + config.defaultParameters.toString());

    if (config.defaultParameters != null) {
      config.defaultParameters!.forEach((key, value) {
        if (!isolateModel.request.has(key)) {
          isolateModel.request.addParam(key: key, value: value);
        }
      });
    }

    String requestMessage = isolateModel.request.toJson();
    _print('raw request : ' + requestMessage);

    requestMessage = _encrypt(crypto, isolateModel, requestMessage);

    return requestMessage;
  }

  String _encrypt(
      Crypto crypto, IsolateModel isolateModel, String requestMessage) {
    if (isolateModel.encrypt != null) {
      if (isolateModel.encrypt!) {
        requestMessage = crypto.encrypt(requestMessage);
        _print('encrypted request : ' + requestMessage);
        return requestMessage;
      }
    } else {
      if (config.encrypt) {
        requestMessage = crypto.encrypt(requestMessage);
        _print('encrypted request : ' + requestMessage);
        return requestMessage;
      }
    }

    return requestMessage;
  }

  String _decrypt(
      Crypto crypto, IsolateModel isolateModel, String responseMessage) {
    if (isolateModel.encrypt != null) {
      if (isolateModel.encrypt!) {
        responseMessage = crypto.decrypt(responseMessage);
        _print('decrypted response : ' + responseMessage);
        return responseMessage;
      }
    } else {
      if (config.encrypt) {
        responseMessage = crypto.decrypt(responseMessage);
        _print('decrypted response : ' + responseMessage);
        return responseMessage;
      }
    }
    return responseMessage;
  }

  Future<Response> makeRequest(IsolateModel isolateModel) async {
    if (await isConnected()) {
      client = http.Client();

      Crypto crypto = Crypto(config.keyPairs?.key);
      String requestMessage = _prepareRequestMessage(crypto, isolateModel);

      if (isolateModel.timeLimit == null)
        isolateModel.timeLimit = config.defaultTimeout;

      http.Response? response = isolateModel.request.method == Method.POST
          ? await _post(isolateModel.url ?? config.url,
              isolateModel.requestParamName, requestMessage,
              encoding: isolateModel.encoding,
              headers: isolateModel.headers,
              timeLimit: isolateModel.timeLimit,
              onTimeOut: isolateModel.onTimeout,
              isPrivate: isolateModel.request.isPrivate)
          : await _get(isolateModel.url ?? config.url,
              isolateModel.requestParamName, requestMessage,
              headers: isolateModel.headers,
              timeLimit: isolateModel.timeLimit,
              onTimeOut: isolateModel.onTimeout);

      if (response != null) {
        if (response.statusCode == 200) {
          // OK
          var responseMessage;
          if ((isolateModel.encrypt ?? false) || config.encrypt) {
            responseMessage = response.body.replaceAll(' ', '');
          } else {
            responseMessage = response.body;
          }
          _print('raw response : ' + responseMessage);

          responseMessage = _decrypt(crypto, isolateModel, responseMessage);

          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print('json decoded : ' + jsonResponse.toString());
            final r = Response(
              response.body,
              response.statusCode,
              result: jsonResponse,
              state: ResponseState.SUCCESS,
            );
            return r;
          } catch (e) {
            _print('json decoded exception : ' + e.toString());
            final r = Response(
              response.body,
              response.statusCode,
              result: {},
              state: ResponseState.UNEXPECTED,
            );
            return r;
          }
        } else {
          // error
          _print(response.statusCode.toString() + ' ' + response.body);
          final r = Response(response.body, response.statusCode,
              result: {}, state: ResponseState.ERROR);
          return r;
        }
      } else {
        _print('http response is null!');
        final r = Response('http response is null!', -1,
            result: {}, state: ResponseState.ERROR);
        return r;
      }
    } else {
      return Response('Check your network connection!', -1,
          result: {}, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  /// This will use [filePath] if [blobData] is null.
  Future<Response> uploadFile(String fileKey, String filePath,
      {String? blobName,
      List<int>? blobData,
      String? url,
      IsolateModel? isolateModel,
      StreamTransformer<List<int>, String>? decoder,
      Function(int bytes, int totalBytes)? onProgress,
      Function(Function c)? cancelToken,
      OnTimeout? onTimeout,
      Duration? timeLimit}) async {
    if (await isConnected()) {
      if (client == null) client = http.Client();
      if (timeLimit == null) timeLimit = config.uploadTimeout;

      var request = isolateModel!.request;
      var req = FileRequest(request.method.value, Uri.parse(url ?? config.url),
          (bytes, totalBytes) {
        _print('Upload progress : ' + (bytes / totalBytes).toString());
        if (onProgress != null) onProgress(bytes, totalBytes);
      });
      if (blobData != null)
        req.files.add(http.MultipartFile.fromBytes(fileKey, blobData,
            contentType: MediaType('application', 'octet-stream'),
            filename: blobName));
      else
        req.files.add(await http.MultipartFile.fromPath(fileKey, filePath));

      Crypto crypto = Crypto(config.keyPairs!.key);

      String requestMessage = _prepareRequestMessage(crypto, isolateModel);

      req.fields[isolateModel.requestParamName] = requestMessage;
      req.fields['os'] = _getOS();

//      cancelToken?.throwIfCancellationRequested();
//      var res = await RetryHelper.tryRun<http.StreamedResponse>(() {
//        return CancellationTokenSource.register(cancelToken, req.send());
//      }, cancelToken: cancelToken);
      if (cancelToken != null) cancelToken(req.close);

      final http.StreamedResponse? res = await req.send().timeout(timeLimit, onTimeout: onTimeout);

      if (res != null) {
        if (res.statusCode == 200) {
          var responseBody = await res.stream.bytesToString();
          var responseMessage = responseBody;
          _print('raw response : ' + responseMessage);

          responseMessage = _decrypt(crypto, isolateModel, responseMessage);

          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print('json decoded : ' + jsonResponse.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: jsonResponse,
              state: ResponseState.SUCCESS,
            );
          } catch (e) {
            _print('json decoded exception : ' + e.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: {},
              state: ResponseState.UNEXPECTED,
            );
          }
//        res.stream
//            .transform(decoder ?? convert.utf8.decoder)
//            .listen((responseBody) {
//          var responseMessage = responseBody;
//
//
//        });
        } else {
          // error
          _print(res.statusCode.toString() + ' ' + res.reasonPhrase!);
          return Response(res.reasonPhrase, res.statusCode,
              result: {}, state: ResponseState.ERROR);
        }
      } else {
        _print('http response is null!');
        return Response('http response is null!', -1,
            result: {}, state: ResponseState.ERROR);
      }
    } else {
      return Response('Check your network connection!', -1,
          result: {}, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  bool isJson(str) {
    try {
      jsonDecode(str);
    } on FormatException {
      return false;
    }
    return true;
  }

  Future<Response> uploadFiles(
      {String? url,
      IsolateModel? isolateModel,
      StreamTransformer<List<int>, String>? decoder,
      List<FileEntity>? multiKey,
      MultipleFileEntity? singleKey,
      List<BlobFileEntity>? multiKeyBlobFile,
      MultipleBlobFileEntity? singleKeyBlobFile,
      Function(int bytes, int totalBytes)? onProgress,
      Function(Function c)? cancelToken,
      OnTimeout? onTimeout,
      Duration? timeLimit}) async {
    if (await isConnected()) {
      if (client == null) client = http.Client();
      if (timeLimit == null) timeLimit = config.uploadTimeout;

      var request = isolateModel!.request;
      var req = FileRequest(request.method.value, Uri.parse(url ?? config.url),
          (bytes, totalBytes) {
        _print('Upload progress : ' + (bytes / totalBytes).toString());
        if (onProgress != null) onProgress(bytes, totalBytes);
      });

      if (singleKeyBlobFile != null) {
        singleKeyBlobFile.blobEntries.forEach((blobEntry) {
          req.files.add(http.MultipartFile.fromBytes(
              singleKeyBlobFile.key, blobEntry.blobData,
              contentType: MediaType('application', 'octet-stream'),
              filename: blobEntry.blobName));
        });
      }

      if (multiKeyBlobFile != null) {
        multiKeyBlobFile.forEach((file) {
          req.files.add(http.MultipartFile.fromBytes(
              file.key, file.blobEntry.blobData,
              contentType: MediaType('application', 'octet-stream'),
              filename: file.blobEntry.blobName));
        });
      }

      if (multiKey != null) {
        multiKey.forEach((file) async {
          req.files.add(await http.MultipartFile.fromPath(file.key, file.path));
        });
      }

      if (singleKey != null) {
        singleKey.paths.forEach((path) async {
          req.files.add(await http.MultipartFile.fromPath(singleKey.key, path));
        });
      }

      Crypto crypto = Crypto(config.keyPairs!.key);
      if (request != null) {
        String requestMessage = _prepareRequestMessage(crypto, isolateModel);

        req.fields[isolateModel.requestParamName] = requestMessage;
        req.fields['os'] = _getOS();
      }

      if (cancelToken != null) cancelToken(req.close);

      var res = await req.send().timeout(timeLimit, onTimeout: onTimeout);

      if (res != null) {
        if (res.statusCode == 200) {
          var responseBody = await res.stream.bytesToString();
          var responseMessage = responseBody;
          _print('raw response : ' + responseMessage);

          responseMessage = _decrypt(crypto, isolateModel, responseMessage);

          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print('json decoded : ' + jsonResponse.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: jsonResponse,
              state: ResponseState.SUCCESS,
            );
          } catch (e) {
            _print('json decoded exception : ' + e.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: {},
              state: ResponseState.UNEXPECTED,
            );
          }
//        res.stream
//            .transform(decoder ?? convert.utf8.decoder)
//            .listen((responseBody) {
//
//        });
        } else {
          // error
          _print(res.statusCode.toString() + ' ' + res.reasonPhrase!);
          return Response(res.reasonPhrase, res.statusCode,
              result: {}, state: ResponseState.ERROR);
        }
      } else {
        _print('http response is null!');
        return Response('http response is null!', -1,
            result: {}, state: ResponseState.ERROR);
      }
    } else {
      return Response('Check your network connection!', -1,
          result: {}, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  Future<http.Response?> _post(
      String url, String requestName, String requestMessage,
      {timeLimit, onTimeOut, headers, encoding, bool isPrivate = true}) async {
    var requestBody = {
      'os': _getOS(),
      'private': (isPrivate ? 1 : 0),
      'version': (isPrivate ? config.privateVersion : config.publicVersion),
      requestName: requestMessage
    };
    return client!
        .post(Uri.parse(url),
            headers: headers,
            body: {'request': convert.jsonEncode(requestBody)},
            encoding: encoding ?? convert.Encoding.getByName('utf-8'))
        .timeout(timeLimit, onTimeout: onTimeOut);
  }

  Future<http.Response> _get(
      String url, String requestName, String requestMessage,
      {timeLimit, onTimeOut, headers}) async {
    String queryUrl =
        url + '?' + requestName + '=' + requestMessage + '&os=' + _getOS();
    return client!
        .get(Uri.parse(queryUrl), headers: headers)
        .timeout(timeLimit, onTimeout: onTimeOut);
  }

  String _getOS() {
    if (kIsWeb) return 'W';

    switch (defaultTargetPlatform) {
      case TargetPlatform.android:
        return 'A';
      case TargetPlatform.iOS:
        return 'I';
      default:
        return 'Unknown';
    }
  }

  void _print(String message) {
    if (config.debugMode) print('[APEX_FLUTTER_PLUGIN] ' + message + '\r\n');
  }

  dispatch() {
    if (client != null) client!.close();
  }

  Future<bool> isConnected() async {
    // TODO connectivity package makes ApexFlutterPlugin not testable :|
    // return true;
    if (kIsWeb) return true;

    var connectivityResult = await (Connectivity().checkConnectivity());
    if (connectivityResult != ConnectivityResult.none) {
      // I am connected to a mobile network.
      return true;
    }
    return false;
  }
}

class FileEntity {
  final String key;
  final String path;

  FileEntity(this.key, this.path);
}

class MultipleFileEntity {
  final String key;
  final List<String> paths;

  MultipleFileEntity(this.key, this.paths);
}

class BlobFileEntity {
  final String key;
  final BlobEntry blobEntry;

  BlobFileEntity(this.key, this.blobEntry);
}

class MultipleBlobFileEntity {
  final String key;
  final List<BlobEntry> blobEntries;

  MultipleBlobFileEntity(this.key, this.blobEntries);
}

class BlobEntry {
  final String blobName;
  final List<int> blobData;

  BlobEntry(this.blobName, this.blobData);
}

class IsolateModel {
  final String requestParamName;
  final Request request;
  final String? url;
  final convert.Encoding? encoding;
  final Map<String, String>? headers;
  final OnTimeout? onTimeout;
  final bool? encrypt;

  Duration? timeLimit;

  IsolateModel(
    this.request, {
    this.requestParamName = 'data',
    this.url,
    this.encoding,
    this.headers,
    this.timeLimit,
    this.onTimeout,
    this.encrypt,
  });
}

typedef OnTimeout = FutureOr<http.StreamedResponse> Function()?;
