import 'dart:convert';

import 'package:apex_flutter_plugin/api/cipher/rsa.dart';
import 'package:apex_flutter_plugin/api/key_pair.dart';

import 'aes.dart';

class Crypto {
  AESAlgorithm? _aes;
  RSAAlgorithm? _rsa;

  Crypto(KeyPair? pair) {
    if (pair?.secretKey != null && pair?.publicKey != null) {
      _aes = AESAlgorithm(pair?.secretKey);
      _rsa = RSAAlgorithm(pair?.publicKey);
    }
  }

  /// Encrypts [message] in my way (combines [_aes] and [_rsa])
  ///
  /// Returns `RSAEncryptedIV:AESEncryptedMessage`
  String encrypt(String message) {
    return _rsa!.encrypt(_aes!.initVector.base64) +
        ":" +
        base64Encode(
          utf8.encode(
            _aes!.encrypt(
              base64Encode(
                utf8.encode(
                  message,
                ),
              ),
            ),
          ),
        );
  }

  /// Decrypts [message] with AES algorithm
  ///
  /// Returns AES decrypted [String]
  String decrypt(String message) {
    return _aes!.decrypt(message);
  }
}
