import 'package:apex_flutter_plugin/api/request.dart';
import 'package:apex_flutter_plugin/server.dart';
import 'package:shared_preferences/shared_preferences.dart';

class LoginRequest extends Request {
  final int countryCode;
  final String username;
  final String password;

  LoginRequest(this.countryCode, this.username, this.password)
      : super({'cc': countryCode, 'username': username, 'password': password},
            isPrivate: true, action: 1002);
}

class LoginAction extends BaseAction {
  final String? token;
  final List<Map<String, dynamic>>? downloadLinks;

  LoginAction(int success, String message, this.token,
      {this.downloadLinks})
      : super(success, message) {
    if (success == 1 && token != null) saveCredentials();
  }

  void saveCredentials() async {
    SharedPreferences prefs = await SharedPreferences.getInstance();
    prefs.setString('apex_api_token', token!);
  }

  factory LoginAction.fromJson(Map<String, dynamic> json) {
    List<Map<String, dynamic>>? download;
    if (json['success'] == -2) {
      download = List<Map<String, dynamic>>.from(json['download']);
    }
    return LoginAction(json['success'], json['message'] ?? '',
        json.containsKey('token') ? json['token'] : null,
        downloadLinks: download);
  }
}
